class Spec:
  def __init__(self, *args, **kwargs):
    pass

  def state_dict(self):
    pass

  def check(self, *args, **kwargs):
    raise NotImplementedError()

  @classmethod
  def infer(cls):
    pass


class ModelSpec(Spec):
  def __init__(self, inputs, outputs, targets):
    super().__init__()


class Field:
  def __init__(
      self,
      name=None,
      description=None,
      required=True,
      default=None,
      validate=None,
      **props):
    self.name = name
    self.description = description
    self.required = required
    self.default = default
    self.validate = validate
    self.__dict__.update(props)


class Tensor(Field):
  pass


class Image(Tensor):
  pass


class Classes(Field):
  pass


class OneHot(Classes):
  pass


Resnet = lambda x: x

Resnet.spec = ModelSpec(
  inputs=Spec(
    Image(
      shape=(3, 224, 224),
      range=(0, 1),
      space='RGB',
      mean=(0.5, 0.5, 0.5),
      std=(0.5, 0.5, 0.5),
      dims=('channel', 'row', 'col'),
      dtype=None,
    )
  ),
  outputs=Spec(

  ),
  targets=Spec(
    OneHot(1000)
  ),
)


def adapt(src: Field, dest: Field):
  pass



import torch

from .utils.registry import Registry, pass_args

## Configure Registry

registry = Registry()

# Datasets
from torchvision import datasets

from torch.utils.data import Dataset

registry.dataset.index(
  datasets,
  types=(Dataset,),
  init=lambda D, root=None, download=True, **kwargs: \
    D(root=root or f'~/.torch/datasets/{D.__name__}',
      download=download)
)

# Losses
registry.loss.index(
  torch.nn.modules.loss,
  types=(torch.nn.modules.loss._Loss,),
  get_names=lambda x:
  (x.__name__, x.__name__[:-len('Loss')]) if x.__name__.endswith('Loss') else (
    x.__name__,)
)

import torch.nn.functional as F

registry.loss.index(
  F,
  include=lambda x: hasattr(x, '__name__') and 'loss' in x.__name__.lower(),
  get_names=lambda x: (x.__name__, x.__name__[:-len('_loss')]) if
  x.__name__.endswith('_loss') else (x.__name__,)
)

from .modules import loss

registry.loss.update((
  F.cross_entropy,
  F.binary_cross_entropy,
  F.binary_cross_entropy_with_logits,
  loss.soft_target_cross_entropy,
  loss.SoftTargetCrossEntopyLoss
))

# Optimizers
from torch.optim import optimizer

registry.optimizer.index(
  torch.optim,
  types=(optimizer.Optimizer,)
)

registry.optimizer['SGD'].init = \
  lambda SGD, params, lr=.01, momentum=.9, weight_decay=1e-4: \
    SGD(params, lr=lr, momentum=momentum, weight_decay=weight_decay)

# LR Schedulers
from torch.optim import lr_scheduler

registry.lr_scheduler.index(
  lr_scheduler,
  types=(lr_scheduler._LRScheduler,)
)

# Models
from torchvision import models


def is_public_callable(x):
  return (
      hasattr(x, '__name__')
      and not x.__name__.startswith('_')
      and callable(x)
  )


registry.model.torchvision.index(
  models,
  init=pass_args,
  include=is_public_callable

)



from collections import OrderedDict
from contextlib import contextmanager

from datetime import datetime


class Timer:
  def __init__(self, name=None, log=False):
    self.tasks = OrderedDict()
    self.name = name
    self.log = log

  def start(self, name, **extra):
    self.tasks[name] = {
      'start': datetime.now(),
      **extra
    }

    if self.log:
      print('Started', name)

  def end(self, name, **extra):
    task = self.tasks.get(name, {})
    task['end'] = datetime.now()
    task = {**extra, **task}
    self.tasks[name] = task

    if self.log:
      print('Ended', name, ', took',
            (task['end'] - task['start']).total_seconds(), 'seconds')

  @contextmanager
  def task(self, name, **extra):
    self.start(name, **extra)
    yield self
    self.end(name, **extra)

  def summary(self):
    d = OrderedDict()
    for name, data in self.tasks.items():
      start = data.get('start')
      end = data.get('end')

      x = {**data}
      if start and end:
        x['seconds'] = (end - start).total_seconds()
      d[name] = x
    return d

  def __enter__(self):
    self.start(self.name)
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.end(self.name)



import torch
from itertools import zip_longest
from torch.utils import data

from .classes import Classes


class Dataset(data.Dataset):
  def state_dict(self):
    return {
      'name': self.__class__.__name__
    }


class ClassificationDataset(Dataset):
  def __init__(self):
    pass


class TransformDataset(Dataset):
  def __init__(self, dataset, transforms):
    super().__init__()
    self.dataset = dataset
    self.transforms = transforms if isinstance(transforms, tuple) else (
      transforms,)

  def __len__(self):
    return len(self.dataset)

  def __getitem__(self, idx):
    return tuple(
      t(x) if t else x
      for (x, t) in zip_longest(self.dataset[idx], self.transforms)
    )

  def __repr__(self):
    return (
      f'{self.__class__.__name__}('
      f'\nDataset: {repr(self.dataset)}'
      f'\nTransforms: {repr(self.transforms)}'
      f'\n)')


class Noisy(Dataset):
  def __init__(self, dataset, p=0.1):
    self.clean_dataset = dataset
    self.p = p

  def __len__(self):
    return len(self.clean_dataset)

  def __getitem__(self, idx):
    x, y = self.clean_dataset[idx]


class Imbalanced(Dataset):
  def __init__(self, dataset):
    self.dataset = dataset


class InputsTargetsDataset(Dataset):
  def __init__(self, inputs, targets, transform=None):
    self.inputs = inputs
    self.targets = targets
    self.transform = transform
    self.classes = Classes(sorted(set(x for t in self.targets for x in t)))

  def __len__(self):
    return len(self.inputs)

  def __getitem__(self, idx):
    x, y = self.inputs[idx], self.targets[idx]
    if self.transform:
      x = self.transform(x)
    return x, y


class OutputCache:
  pass


class TinyDigits(data.TensorDataset):
  """
  Dataset of 8x8 digits, best used for testing
  """

  def __init__(self, num_classes=10):
    from sklearn.datasets import load_digits
    digits = load_digits(num_classes)
    super().__init__(
      torch.from_numpy(digits.images).unsqueeze(1).float(),
      torch.Tensor(digits.target).long()
    )



import json
import os
import pickle as pkl
import tarfile

import numpy as np
import re
import torch
from PIL import Image


def camel_to_snake(text):
  s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', text)
  return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def truthy(items):
  return [x for x in items if x]


class Obj(dict):
  __setattr__ = dict.__setitem__
  __getattr__ = dict.__getitem__


def save_pickle(obj, path, mode='wb'):
  with open(str(path), mode) as f:
    pkl.dump(obj, f)


def load_pickle(path, mode='rb'):
  with open(str(path), mode) as f:
    return pkl.load(f)


def save_json(obj, path, mode='w'):
  with open(str(path), mode) as f:
    json.dump(obj, f)


def load_json(path, mode='r'):
  with open(str(path), mode) as f:
    return json.load(f)


def tar_dir(path, dest=None, zip=True):
  path = str(path)
  dest = str(dest or path)

  ext, mode = ('.tar.gz', 'w:gz') if zip else ('.tar', 'w')

  if not dest.endswith(ext):
    dest = os.path.splitext(dest)[0] + ext

  with tarfile.open(dest, mode) as tar:
    tar.add(path)


def untar(path):
  with tarfile.open(path) as tar:
    tar.extractall()


def almost_equal(t1, t2, prec=1e-12):
  return torch.all(torch.lt(torch.abs(torch.add(t1, -t2)), prec))


def equal(t1, t2):
  return torch.all(t1 == t2)


def randimg(*shape):
  return Image.fromarray((np.random.rand(*shape) * 255).astype('uint8'))
